{$IFNDEF FPC}
  {$N+}
  {$E+}
{$ENDIF}
{
              Program Name : Maths
              Package Name : LB
              Purpose      : Various math routines
              Institution  : Conroy & Conroy Company
              Original Vers: 01-Apr-1987
              Date Written : 25-Jan-2002
              Written By   : Alan Conroy
              Version      : 2.0C

              Copyright (C) 1987-2016 by Alan Conroy.  Released to the public domain.

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

             DATE      BY             REASON

          4-Sep-2002   EAC            Added GCD.

          1-Nov-2002   EAC            Allow operation on both
                                      systems with and without FPUs.
                                      Added numerous other math functions.
          20-Jan-2003  EAC            Added Min and Max.
          28-Feb-2016  EAC            Handle factorials of integers better.
          27-Feb-2020  EAC            Add Variance and Standard_Deviation.

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          These functions perform various math functions.
}

unit Maths ;

{$I EDEFINES.INC}

interface

// General...
function Max( X, Y : integer ) : integer ;
function Min( X, Y : integer ) : integer ;
{$IFDEF INT64}
function Round_Up( Value, Round : int64 ) : int64 ;
{$ELSE}
function Round_Up( Value, Round : cardinal ) : cardinal ;
{$ENDIF}
{$IFDEF INT64}
function Convert_Endian( Value : int64 ; Size : integer ) : int64 ;
{$ELSE}
function Convert_Endian( Value : cardinal ; Size : integer ) : cardinal ;
{$ENDIF}

function Root( X : Extended ; R : longint ) : Extended ;
function Sgn( X : real ) : real ;
function ISgn( X : longint ) : integer ;
function Power( X, N : Extended ; var _Result : Integer ) : Extended ;
function GCD( N, D : longint ) : longint ;
function Factorial( X : extended ) : extended ;
function SuperFactorial( X : extended ) : extended ;
function Circle_Area( Radius : extended ) : extended ;
function Circle_Circumference( Diameter : extended ) : extended ;

// Constants...
function E : extended ;
function Pi : Extended ;

// Trigonometric...
function Tan( X : Extended ) : Extended ; { Return tangent of a value }
function Log10( X : Extended ) : Extended ; { Return logarithm, base 10 }
function Rad( X : extended ) : extended ; { Degree to Radian conversion }
function Deg( X : extended ) : extended ; { Radian to Degree conversion }
function CoT( X : extended ) : extended ; { Cotangent }
function CSc( X : extended ) : extended ; { Cosecant }
function Sec( X : extended ) : extended ; { Secant }
function Fib( X : extended ) : extended ; { Fibonacci sequence }
function ArcCos( X : extended ) : extended ; { Arc Cosine }
function ArcCoT( X : extended ) : extended ; { Arc Cotangent }
function ArcCSc( X : extended ) : extended ; { Arc Cosecant }
function ArcSec( X : extended ) : extended ; { Arc Secant }
function ArcSin( X : extended ) : extended ; { Arc Sine }
function HypCos( X : extended ) : extended ; { Hyperbolic Cosine }
function HypCoT( X : extended ) : extended ; { Hyperbolic Cotangent }
function HypCSc( X : extended ) : extended ; { Hyperbolic Cosecant }
function HypSec( X : extended ) : extended ; { Hyperbolic secant }
function HypSin( X : extended ) : extended ; { Hyperbolic Sine }
function HypTan( X : extended ) : extended ; { Hyperbolic Tangent }
function HypArcCos( X : extended ) : extended ; { Hyperbolic Arc Cosine }
function HypArcCoT( X : extended ) : extended ; { Hyperbolic Arc Cotangent }
function HypArcCsc( X : extended ) : extended ; { Hyperbolic Arc Cosecant }
function HypArcSec( X : extended ) : extended ; { Hyperbolic Arc Secant }
function HypArcSin( X : extended ) : extended ; { Hyperbolic Arc Sine }
function HypArcTan( X : extended ) : extended ; { Hyperbolic Arc Tangent }

// Geometry...
function Slope( X, Y : extended ) : extended ;
function X_From_Slope( S : extended ) : extended ;
function Y_From_Slope( S : extended ) : extended ;
procedure Pythagorean_Triple( Slope : extended ; var a, b, c : extended ) ;
function Euclidean_Distance( const Point1, Point2 : array of longint ) : extended ;
function Manhattan_Distance( const Point1, Point2 : array of longint ) : extended ;

// Statistical..
function Benford( X : extended ) : extended ; // Benford's distribution
function Variance( const Values : array of extended) : extended ;
function Standard_Deviation( const Values : array of extended ) : extended ;
function Mean( const Values : array of extended ) : extended ;
function Sum( const Values : array of extended ) : extended ;
function Coefficient_of_variation( const Values : array of extended ) : extended ;
function Probability( X, Mean, Standard_Deviation : extended ) : extended ;
function Range_Probability( X, Y, Min, Max : extended ) : extended ;
function Choose( N, K : extended ) : extended ;
function Collision_Probability( Tries, Size : cardinal ) : extended ;
function SEM( std : extended ; Sample_Size : cardinal ) : extended ; // Standard Error of the Mean
function Coefficient_of_Determination_Linear( const Observed, Predicted : array of extended ) : extended ;

// Classifiers...
function Sensitivity( True_Positive, False_Negative : extended ) : extended ;
function Specificity( True_Negative, False_Positive : extended ) : extended ;
function Predictive_Value( True_Value, False_Value : extended ) : extended ;
function Odds( Probability : extended ) : extended ;
function Probability_From_Odds( Odds : extended ) : extended ;


implementation

uses { Subroutine Library - Layer 0... }
     _Types, { Small_String }
     L10n, { Text_* }

     { Subroutine Library - Layer 1 }
     _Applic, { Application_Manager }
     CommonUt ; { Edit }


function Max( X, Y : integer ) : integer ;

begin
    if( X > Y ) then
    begin
        Max := X ;
    end else
    begin
        Max := Y ;
    end ;
end ;


function Min( X, Y : integer ) : integer ;

begin
    if( X < Y ) then
    begin
        Min := X ;
    end else
    begin
        Min := Y ;
    end ;
end ;


{$IFDEF INT64}
function Round_Up( Value, Round : int64 ) : int64 ;

begin
    if( Round = 0 ) then
    begin
        Result := 0 ;
        exit ; 
    end ;
    Result := ( Value + Round - 1 ) div Round ;
    Result := Result * Round ;
end ;
{$ELSE}
function Round_Up( Value, Round : cardinal ) : cardinal ;

begin
    if( Round = 0 ) then
    begin
        Result := 0 ;
        exit ; 
    end ;
    Result := ( Value + Round - 1 ) div Round ;
    Result := Result * Round ;
end ;
{$ENDIF}


{$IFDEF INT64}
function Convert_Endian( Value : int64 ; Size : integer ) : int64 ;

var A : array[ 0..7 ] of byte ;
    Loop, Z : integer ;

begin
    Result := 0 ;
    move( Value, A, sizeof( Value ) ) ;
    for Loop := 0 to sizeof( A ) div 2  - 1 do
    begin
        Z := A[ Loop ] ;
        A[ Loop ] := A[ sizeof( A ) - Loop - 1 ] ;
        A[ sizeof( A ) - Loop - 1 ] := Z ;
    end ;
    move( A[ 8 - Size ], Result, Size ) ;
end ;
{$ELSE}
function Convert_Endian( Value : cardinal ; Size : integer ) : cardinal ;

var A : array[ 0..3 ] of byte ;
    Loop, Z : integer ;

begin
    Result := 0 ;
    move( Value, A, sizeof( Value ) ) ;
    for Loop := 0 to sizeof( A ) div 2  - 1 do
    begin
        Z := A[ Loop ] ;
        A[ Loop ] := A[ sizeof( A ) - Loop - 1 ] ;
        A[ sizeof( A ) - Loop - 1 ] := Z ;
    end ;
    move( A[ 4 - Size ], Result, Size ) ;
end ;
{$ENDIF}


function SuperFactorial( X : extended ) : extended ;

var Y : extended ;

begin
    Y := X ;
    while( Y > 1 ) do
    begin
        X := X * Factorial( Y ) ;
        Y := Y - 1 ;
    end ;
    SuperFactorial := X ;
end ;


function Factorial( X : extended ) : extended ;

var Y : extended ;

begin
    Y := X - 1 ;
    if( X = int( X ) ) then
    begin
        while( Y > 1 ) do
        begin
            X := int( X * Y ) ;
            Y := Y - 1 ;
        end ;
    end else
    begin
        while( Y > 1 ) do
        begin
            X := X * Y ;
            Y := Y - 1 ;
        end ;
    end ;
    Factorial := X ;
end ;


function E : extended ;

begin
    E := 2.718 ;
end ;


function Pi : Extended ;

begin
    Pi := 3.1415926535897932385 ;
end ;


function Circle_Area( Radius : extended ) : extended ;

begin
    Circle_Area := PI * SQR( Radius ) ;
end ;


function Circle_Circumference( Diameter : extended ) : extended ;

begin
    Circle_Circumference := Pi * Diameter ;
end ;


function Tan( X : Extended ) : Extended ; { Return tangent of a value }

begin
    Tan := Sin( X ) / Cos( X )
end ;


function Log10( X : Extended ) : Extended ; { Return logarithm, base 10 }

begin
    Log10 := Ln( X ) / Ln( 10 )
end ;


function Rad( X : extended ) : extended ; { RAD - Degree to Radian conversion }

begin
    Rad := X / ( 180. / Pi ) ;
end ;


function Deg( X : extended ) : extended ; { DEG - Radian to Degree conversion }

begin
    Deg := X * ( 180. / Pi ) ;
end ;


function CoT( X : extended ) : extended ; { COT - Cotangent }

begin
    CoT := 1 / Tan( X ) ;
end ;


function CSc( X : extended ) : extended ; { CSC - Cosecant }

begin
    CSc := 1. / Sin( X ) ;
end ;


function Sec( X : extended ) : extended ; { SEC - Secant }

begin
    Sec := 1. / Cos( X ) ;
end ;


function Fib( X : extended ) : extended ; { Fibonacci sequence }

var A1, A2, A3 : extended ;
    Dummy : integer ;

begin
    X := Int( Abs( X ) ) ;
    if( X < 2 ) then
    begin
        Fib := 1 ;
    end else
    begin
        A1 := 1 ;
        A2 := 1 ;
        for Dummy := 2 to Trunc( X ) do
        begin
            A3 := A2 ;
            A2 := A1 ;
            A1 := A2 + A3 ;
        end ;
        Fib := A1 + A2 ;
    end
end ;


function GCD( N, D : longint ) : longint ;
{ Return greatest common divisor of passed numbers }

var X : longint ;

begin
    if( N = D ) then
    begin
	GCD := D ;
	exit ;
    end ;
    if( N < D ) then
    begin
	GCD := 0 ; { Greatest common divisor is non-integer }
	exit ;
    end ;
    X := N mod D ;
    while( X <> 0 ) do
    begin
	N := D ;
	D := X ;
	X := N mod D ;
    end ;
    GCD := D ;
end ;


function ArcCos( X : extended ) : extended ; { Arc Cosine }

var E : integer ;
    Temp : extended ;

begin
    Temp := Power( X, 2, E ) ;
    Temp := 1 - Temp ;
    ArcCos := Arctan( Sqrt( Temp ) / X )
end ;


function ArcCoT( X : extended ) : extended ; { Arc Cotangent }

begin
    ArcCoT := Arctan( 1. / X )
end ;


function ArcCSc( X : extended ) : extended ; { Arc Cosecant }

begin
    ArcCSc := Arctan( 1. / Sqrt( X ) ) ;
end ;


function ArcSec( X : extended ) : extended ; { Arc Secant }

var E : integer ;

begin
    X := Power( X, 2, E ) - 1 ;
    ArcSec := Arctan( Sqrt( X ) ) ;
end ;


function ArcSin( X : extended ) : extended ; { Arc Sine }

var E : integer ;
    Temp : extended ;

begin
    Temp := Power( X, 2, E ) ;
    Temp := 1 - Temp ;
    ArcSin := Arctan( X / Sqrt( Temp ) )
end ;


function HypCos( X : extended ) : extended ; { Hyperbolic Cosine }

begin
    HypCos := ( Exp( X ) + Exp( - X ) ) / 2.
end ;


function HypCoT( X : extended ) : extended ; { Hyperbolic Cotangent }

var Temp : extended ;

begin
    Temp := Exp( X ) - Exp( - X ) ;
    HypCoT := ( Exp( X ) + Exp( -X ) ) / Temp ;
end ;


function HypCSc( X : extended ) : extended ; { Hyperbolic Cosecant }

begin
    X := Exp( X ) - Exp( -X ) ;
    HypCSc := 2. / X ;
end ;


function HypSec( X : extended ) : extended ; { Hyperbolic secant }

begin
    X := Exp( X ) + Exp( -X ) ;
    HypSec := 2. / X ;
end ;


function HypSin( X : extended ) : extended ; { Hyperbolic Sine }

begin
    HypSin := ( Exp( X ) - Exp( -X ) ) / 2.
end ;


function HypTan( X : extended ) : extended ; { Hyperbolic Tangent }

var Temp : extended ;

begin
    Temp := Exp( X ) + Exp( -X ) ;
    HypTan := ( Exp( X ) - Exp( -X ) ) / Temp ;
end ;


function HypArcCos( X : extended ) : extended ; { Hyperbolic Arc Cosine }

var E : integer ;
    Temp : extended ;

begin
    Temp := Power( X, 2, E ) ;
    Temp := Temp - 1 ;
    X := X + Sqrt( Temp ) ;
    HypArcCos := Ln( X )
end ;


function HypArcCoT( X : extended ) : extended ; { Hyperbolic Arc Cotangent }

begin
    HypArcCoT := ( Ln( X + 1 ) - Ln( X - 1 ) ) / X ;
end ;


function HypArcCsc( X : extended ) : extended ; { Hyperbolic Arc Cosecant }

var E : integer ;
    Temp : extended ;

begin
    Temp := Power( X, 2, E ) ;
    HypArcCsc := Ln( ( 1. / X ) + Sqrt( 1 / Temp + 1 ) )
end ;


function HypArcSec( X : extended ) : extended ; { Hyperbolic Arc Secant }

var E : integer ;
    Temp : extended ;

begin
    Temp := Power( X, 2, E ) ;
    Temp := 1 / Temp - 1 ;
    HypArcSec := Ln( ( 1. / X ) + Sqrt( Temp ) ) ;
end ;


function HypArcSin( X : extended ) : extended ; { Hyperbolic Arc Sine }

var E : integer ;
    Temp : extended ;

begin
    Temp := Power( X, 2, E ) + 1 ;
    X := X + Sqrt( Temp ) ;
    HypArcSin := Ln( X )
end ;


function HypArcTan( X : extended ) : extended ; { Hyperbolic Arc Tangent }

begin
    HypArcTan := ( Ln( ( 1. + X ) / ( 1. - X ) ) ) / 2. ;
end ;


function Root( X : Extended ; R : longint ) : Extended ;
{ This function returns the Rth root of X.  For example, Root( 27, 3 ) returns 3. }

var Dummy : integer ;
    Last, Low, High, Mid, Intermediate : Extended ;

begin
    { Check invalid results }
    if( X < 0 ) then { Negative value }
    begin
        if( ( R and 1 ) = 0 ) then { Even number }
        begin
            Application_Manager^.Fatal_Exception( Text_Error_Imaginary_Number ) ;
            { Cannot get a negative value from an even number of multiplies }
        end ;
    end ;

    { Determine initial range }
    High := R ;
    High := X / High ;
    if( High < 2 ) then
    begin
	      High := 2 ;
    end ;
    Low := 1.0 ;
    Last := X ;

    { Bisearch for the solution }
    while( Low <> High ) do
    begin
        Mid := Low + ( ( High - Low ) / 2.0 ) ;
        if( ( Low = Mid ) or ( High = Mid ) ) then
        begin
            break ;
        end ;
        Intermediate := Power( Mid, R, Dummy ) ;
        if( Intermediate = X ) then { Match found }
        begin
            Low := Mid ;
            break ;
        end else
        begin
            if( Intermediate < X ) then
            begin
                Low := Mid ;
            end else
            begin
                High := Mid ;
            end ;
        end ;
        if( Last = Intermediate ) then { No further resolution available }
        begin
            Low := Mid ;
            break ;
        end ;
        Last := Intermediate ;
    end ; { while( Low <> High ) }

    { Return result }
    Root := Low ;
end ; { Root }


function Sgn( X : Real ) : real ;
{This function emulates the BASIC-Plus SGN function.  It returns the following:
           for SGN(n):
              n<0 returns -1
              n=0 returns 0
              n>0 returns 1
}

begin
    if (X = 0 ) then
    begin
        Sgn := 0 ;
    end else
    if( X < 0 ) then
    begin
        Sgn := -1 ;
    end else
    begin
        Sgn := 1 ;
    end ;
end ;


function ISgn( X : longint ) : integer ;
{ This function emulates the BASIC-Plus SGN function.  It returns the following:
           for SGN(n):
              n<0 returns -1
              n=0 returns 0
              n>0 returns 1
}

begin
    if( X = 0 ) then
    begin
        ISgn := 0 ;
    end else
    if( X < 0 ) then
    begin
        ISgn := -1 ;
    end else
    begin
        ISgn := 1 ;
    end ;
end ;


function Power( X, N : Extended ; var _Result : Integer ) : Extended ;
{ This function returns the result of one number to the power of another.
  Based on C code from Computer Language magazine. }

var Size_X : Integer ;
    Temp : Small_string ;

    function Pwr( N : Integer ) : Extended ;

    var Temp1 : Real ;

    begin
        Pwr := -1 ; { Assume failure }
        if( _Result <> 0 ) then
        begin
            Pwr := 1 ;
            Exit
        end ;
        if( N = 1 ) then
        begin
            Pwr := X ;
        end else
        begin
            Temp1 := Pwr( N div 2 ) ;
            Str( Temp1 : 40 : 0, Temp ) ;
            Temp := Edit( Temp, 2 ) ;
            if( Odd( N ) ) then
            begin
                if( Size_X + Length( Temp ) * 2 < 32 ) then
                begin
                    Pwr := Sqr( Temp1 ) * X ;
                end else
                begin
                    _Result := 7 ;
                    Exit ;
                end
            end else
            if( Length( Temp ) * 2 < 32 ) then
            begin
                Pwr := Sqr( Temp1 ) ;
            end else
            begin
                _Result := 7 ;
            end ;
        end ; { if( N = 1 ) }
    end ; { Power.Pwr }

begin { Power }
    _Result := 0 ;
    Power := 0 ;
    if( X <> 0 ) then
    begin
        if( N = 0 ) then
        begin
            Power := 1 ;
        end else
        if( Frac( N ) <> 0 ) then
        begin
            if ( X < 0 ) then
            begin
                _Result := 11 ;
                Exit ;
            end ;
            if ( N * Ln( X ) > 88 ) or ( N * Ln( X ) < -88 ) then
            begin
                _Result := 12 ; { Error }
                Exit ;
            end ;
            Power := Exp( N * Ln( X ) )
        end else
        begin
            Str( X : 40 : 0, Temp ) ;
            Temp := Edit( Temp, 2 ) ;
            Size_X := Length( Temp ) ;
            if( ( N > 32767 ) or ( N < -32767 ) ) then
            begin
                _Result := 7 ;
            end else
            if( N < 0 ) then
            begin
                Power := 1 / Pwr( -Trunc( N ) ) ;
            end else
            begin
                Power := Pwr( Trunc( N ) ) ;
            end ;
        end ; { if( Frac( N ) <> 0 }
    end ; { if( X <> 0 ) }
end ; { Power }


function Slope( X, Y : extended ) : extended ;

begin
    Result := Y / ( X + 1.0 ) ;
end ;


function X_From_Slope( S : extended ) : extended ;

begin
    Result := ( 1.0 - sqr( S ) ) / ( 1.0 + sqr( S ) ) ;
end ;


function Y_From_Slope( S : extended ) : extended ;

begin
    Result := ( 2.0 * S ) / ( 1.0 + sqr( S ) ) ;
end ;


procedure Pythagorean_Triple( Slope : extended ; var a, b, c : extended ) ;

var p, q : extended ;

begin
    p := Y_From_Slope( Slope ) ;
    q := X_From_Slope( Slope ) + 1.0 ;
    a := sqr( q ) - sqr( p ) ;
    b := 2.0 * p * q ;
    c := sqr( q ) + sqr( p ) ;
end ;


function Benford( X : extended ) : extended ;

begin
    Benford := Log10( 1 + 1 / X ) ;
end ;


function Variance( const Values : array of extended ) : extended ;

var I : integer ;
    Mean, Sum, Total : extended ;

begin
    Sum := 0.0 ;
    for I := low( Values ) to high( Values ) do
    begin
        Sum := Sum + Values[ I ] ;
    end ;
    Mean := Sum / length( Values ) ;

    Total := 0.0 ;
    for I := low( Values ) to high( Values ) do
    begin
        Total := Total + SQR( Values[ I ] - Mean ) ;
    end ;

    Variance := Total / length( Values ) ;
end ;


function Standard_Deviation( const Values : array of extended ) : extended ;

begin
    Standard_Deviation := SQRT( Variance( Values ) ) ;
end ;


function Sum( const Values : array of extended ) : extended ;

var I : integer ;
    S : extended ;

begin
    S := 0.0 ;
    for I := low( Values ) to high( Values ) do
    begin
        S := S + Values[ I ] ;
    end ;
    Sum := S ;
end ;


function Mean( const Values : array of extended ) : extended ;

begin
    Mean := Sum( Values ) / length( Values ) ;
end ;


function Coefficient_of_variation( const Values : array of extended ) : extended ;

begin
    Coefficient_of_variation := Standard_Deviation( Values ) / Mean( Values ) ;
end ;


// Probability of X within a normal (Gaussian) distribution
function Probability( X, Mean, Standard_Deviation : extended ) : extended ;

var Dummy : integer ;

begin
    Probability := ( 1 / Standard_Deviation * SQRT( 2 * Pi ) ) * Power( E, SQR( X - Mean ) / SQR( 2 * Standard_Deviation ), Dummy ) ;
end ;


// Possibility of a number in the range of X to Y in a uniform (rectangular) distribution
function Range_Probability( X, Y, Min, Max : extended ) : extended ;

begin
    if( ( X < Min ) or ( Y > Max ) or ( Max = Min ) ) then
    begin
        Range_Probability := 0 ;
        exit ;
    end ;
    Range_Probability := ( Y - X ) / ( Max - Min ) ;
end ;


// Binomial coefficient (n choose k)
function Choose( N, K : extended ) : extended ;

begin
    Choose := Factorial( N ) / ( Factorial( K ) * Factorial( N - K ) )
end ;


// Chance of at least 1 collision after Tries tries in a set of Size length
function Collision_Probability( Tries, Size : cardinal ) : extended ;

var I : cardinal ;
    N, Res : extended ;

begin
    N := Tries ;
    Res := 1.0 ;
    for I := 1 to Tries do
    begin
        Res := Res * ( ( N - I ) / N ) ;
    end ;
    Collision_Probability := 1 - Res ;
end ;


function SEM( std : extended ; Sample_Size : cardinal ) : extended ; // Standard Error of the Mean

begin
    SEM := std / SQRT( Sample_Size ) ;
end ;


function Coefficient_of_Determination_Linear( const Observed, Predicted : array of extended ) : extended ;

var Estimate_Error : extended ;
    I : integer ;
    U : extended ; // Mu (mean)
    Variability : extended ;

begin
    U := Mean( Observed ) ;
    I := length( Observed ) ;
    if( I > length( Predicted ) ) then
    begin
        I := length( Predicted ) ;
    end ;
    Estimate_Error := 0.0 ;
    Variability := 0.0 ;
    while( I > 0 ) do
    begin
        Estimate_Error := Estimate_Error + SQR( Observed[ I ] - Predicted[ I ] ) ;
        Variability := Variability + SQR( Observed[ I ] - U ) ;
    end ;
    Coefficient_of_Determination_Linear := 1 - Estimate_Error / Variability ;
end ;


function Euclidean_Distance( const Point1, Point2 : array of longint ) : extended ;

var I : integer ;
    Sum : longint ;

begin
    Euclidean_Distance := 0.0 ;
    if( ( low( Point1 ) <> low( Point2 ) ) or ( high( Point1 ) <> high( Point2 ) ) ) then
    begin
        exit ; // arrays have different dimensions - cannot calculate
    end ;
    Sum := 0 ;
    for I := low( Point1 ) to high( Point1 ) do
    begin
        Sum := Sum + trunc( SQR( abs( Point1[ I ] - Point2[ I ] ) ) ) ;
    end ;
    if( Sum = 0 ) then
    begin
        exit ;
    end ;
    Euclidean_Distance := SQRT( Sum ) ;
end ;


function Manhattan_Distance( const Point1, Point2 : array of longint ) : extended ;

var I : integer ;
    Sum : longint ;

begin
    Manhattan_Distance := 0.0 ;
    if( ( low( Point1 ) <> low( Point2 ) ) or ( high( Point1 ) <> high( Point2 ) ) ) then
    begin
        exit ; // arrays have different dimensions - cannot calculate
    end ;
    Sum := 0 ;
    for I := low( Point1 ) to high( Point1 ) do
    begin
        Sum := Sum + trunc( abs( Point1[ I ] - Point2[ I ] ) ) ;
    end ;
    Manhattan_Distance := Sum ;
end ;


function Sensitivity( True_Positive, False_Negative : extended ) : extended ;

begin
    Sensitivity := True_Positive / ( True_Positive + False_Negative ) ;
end ;


function Specificity( True_Negative, False_Positive : extended ) : extended ;

begin
    Specificity := True_Negative / ( True_Negative + False_Positive ) ;
end ;


function Predictive_Value( True_Value, False_Value : extended ) : extended ;

begin
    Predictive_Value := True_Value / ( True_Value + False_Value ) ;
end ;


function Odds( Probability : extended ) : extended ;

begin
    Odds := Probability / ( 1 - Probability ) ;
end ;


function Probability_From_Odds( Odds : extended ) : extended ;

begin
    Probability_From_Odds := Odds / ( Odds + 1 ) ;
end ;



end.

