{
        Program Name : UOS_File_Heap
        Package Name : Subroutine Library - Layer 1
        Purpose      : UOS File Heap
        Institution  : Conroy & Conroy Co.
        Date Written : 2017
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2017 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TUOS_File_Heap is a managed store implementing a heap in a UOS file.
}

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

unit UOS_File_Heap ;

interface

uses // Subroutine Library...
     _Stores, // TStore_Address64 ;
     _UE, // TUnified_Exception

     // UOS...
     UOS_File_Store, // TUOS_Native_File_Store
     _UOS_File_System, // TUOS_File_System
     UOS_MStore ; // TUOS_Managed_Store

type TUOS_File_Heap = class( TUOS_Managed_Store )
                          public // Constructors and destructors...
                              constructor Create ;
                              destructor Destroy ; override ;

                          private // Instance data...
                              ATO_File_Position : int64 ; // Location of AT offset pointer in the file
                              _File_Store : TUOS_Native_File_Store ;
                              _Name : string ;
                              _Resolution : longint ;

                          protected // Internal utility routines...
                              function Valid_Range( Root, Length : int64 ) : boolean ;
                              function Get_Length( Root : int64 ) : int64 ;
                              function Init_Heap : TUnified_Exception ;
                              function Round_Up( S, Resolution : longint ) : longint ;

                          protected // Property handlers...
                              function Get_Resolution : longint ;
                              procedure Set_Resolution( Value : longint ) ;
                              function Get_File : TUOS_File ;
                              procedure Set_File( Value : TUOS_File ) ;
                              function Get_Origin : int64 ;
                              procedure Set_Origin( Value : int64 ) ;
                              function Get_Fixed : boolean ;
                              procedure Set_Fixed( Value : boolean ) ;

                          public { Overrides... }
                              function Is_Class( Name : PChar ) : boolean ;
                                  override ;

                              function Read_Data( var Data ; Address,
                                  _Size : TStore_Address64 ;
                                  var UEC : TUnified_Exception ) : TStore_Address64 ;
                                  override ;

                              function Write_Data( var Data ; Address,
                                  _Size : TStore_Address64 ;
                                  var UEC : TUnified_Exception ) : TStore_Address64 ;
                                  override ;

                              function Get_Name : PChar ; override ;
                              { Returns name of store. }

                              procedure Set_Max_Storage( Value : TStore_Address64 ;
                                  var Res : TUnified_Exception ) ;
                                  override ;

                              function Allocate( Size : TStore_Address64 ) :
                                  TStore_Address64 ; override ;
                                  { Allocate space}
                              function Allocate_At( Offset,
                                  Size : TStore_Address64 ) : boolean ;
                                  override ;
                                  { Allocate space at specific location }
                              procedure Copy( Source,
                                  Destination : TStore_Address64 ;
                                  Count : int64 ) ; override ; { Copy data }
                              procedure Deallocate( PTR,
                                  Size : TStore_Address64 ) ;
                                  override ;
                              procedure Fill( PTR : TStore_Address64 ;
                                  Count : integer ; Value : byte ) ;
                                  override ;
                              function Min_Storage : TStore_Address64 ;
                                  override ;
                              function Reallocate( PTR, Old,
                                  New : TStore_Address64 ) : TStore_Address64 ;
                                  override ;

                          public // API...
                              procedure Assign_File( Value : TUOS_File ;
                                  Resolution : integer = 16 ;
                                  Initialize : boolean = False ) ;
                              function Open( FS : TUOS_File_System ;
                                  Name : string ) : TUnified_Exception ;
                                  virtual ;
                              function Getmem( Size : int64 ) : int64 ; virtual ;
                              procedure Freemem( PTR : int64 ) ; virtual ;
                              function Reallocmem( PTR, Size : int64 ) : int64 ;
                                  virtual ;
                              procedure Reset ;

                          public // Properties...
                              property _File : TUOS_File
                                  read Get_File
                                  write Set_File ;
                              property Fixed : boolean
                                  read Get_Fixed
                                  write Set_Fixed ;
                              property Origin : int64
                                 read Get_Origin
                                 write Set_Origin ;
                              property Resolution : longint
                                  read Get_Resolution
                                  write Set_Resolution ;
                      end ; { TUOS_File_Heap }

const Err_UOS_File_Heap_Facility = 144 ;
const Err_Illegal_Call = 1 ;
const Err_Access_Error = 2 ;
const Err_Invalid_Heap_File = 3 ;
const Err_Incompatible_Version = 4 ;


type TFH_Header = packed record
                      Prefix : byte ;
                      Facility : byte ;
                      Version : byte ;
                      Reservedb : byte ;
                      Resolution : longint ;
                      Flags : longint ;
                      Reserved : longint ;
                      AT_Offset : int64 ;
                      Origin : int64 ;
                  end ;

implementation

uses // Delphi...
     Sysutils, // lowercase

     // Subroutine Library...
     CCStandard,
     UE ; // Create_Simple_UE


// TUOS_File_Heap methods...

// Constructors and destructors...

constructor TUOS_File_Heap.Create ;

begin
    inherited Create ;

    _Resolution := 16 ;
end ;


destructor TUOS_File_Heap.Destroy ;

var FH_Header : TFH_Header ;

begin
    if( _File_Store <> nil ) then
    begin
        if( AT <> nil ) then
        begin
            _File_Store._File.Read( 0, 0, sizeof( FH_Header ), FH_Header ) ;
            if( FH_Header.AT_Offset <> AT.Offset ) then
            begin
                FH_Header.AT_Offset := AT.Offset ;
                _File_Store._File.Write( 0, 0, sizeof( FH_Header ), FH_Header ) ;
            end ;
        end ;
        //_File_Store.Detach ;
        Set_Store( nil ) ;
        _File_Store := nil ;
    end ;

    inherited Destroy ;
end ;


// Internal utility routines...

function TUOS_File_Heap.Valid_Range( Root, Length : int64 ) : boolean ;

begin
    Result := ( not ( ( Root < Allocation_Table_Offset + AT.Size ) and ( Root + Length >= Allocation_Table_Offset ) ) ) // AT
              and
              ( not ( Root < Resolution ) ) ; // Cluster 0
end ;


function TUOS_File_Heap.Get_Length( Root : int64 ) : int64 ;

var UEC : TUnified_Exception ;

begin
    inherited Read_Data( Result, Root - sizeof( int64 ), sizeof( int64 ), UEC ) ;
end ;


function Power_Of_2( X : int64 ) : boolean ;

var I : integer ;

begin
    Result := True ; // Assume true
    for I := 0 to 63 do
    begin
        if( XBits_For_Value( I ) = X ) then
        begin
            exit ;
        end ;
    end ;
    Result := False ;
end ;


procedure TUOS_File_Heap.Reset ;

var FH_Header : TFH_Header ;

begin
    fillchar( FH_Header, sizeof( FH_Header ), 0 ) ;
    FH_Header.Prefix := $FF ;
    FH_Header.Facility := Err_UOS_File_Heap_Facility ;
    FH_Header.Resolution := Resolution ;
    Allocation_Table_Offset := Round_Up( sizeof( FH_Header ), Resolution ) ;
    FH_Header.AT_Offset := Allocation_Table_Offset ;
    _File_Store._File.Write( 0, 0, sizeof( FH_Header ), FH_Header ) ;
    AT.Offset := Allocation_Table_Offset ;
    AT.Store := _File_Store ;
    AT.Size := ( _File_Store._File.File_Size div Resolution + 7 ) div 8 ;
    AT.Resolution := Resolution ;
    AT.Allocate( Allocation_Table_Offset + AT.Size + Resolution ) ; // Mark header and AT as allocated
    AT.Flush ;
end ;


function TUOS_File_Heap.Init_Heap : TUnified_Exception ;

var FH_Header : TFH_Header ;
    I : int64 ; // AT Size
    S : int64 ; // New file size

begin
    // Initialize data...
    Result := nil ;
    Set_Last_Error( nil ) ; // Clear errors

    if( _File_Store._File.File_Size = 0 ) then
    begin
        // Initialize new heap...
        Reset ;
        exit ;
    end ;

    // Read the file header...
    if( _File_Store._File.File_Size < sizeof( FH_Header ) ) then // File is smaller than the file header
    begin // UNTESTED
        Result := Create_Simple_UE( Err_UOS_File_Heap_Facility, 10, Err_Invalid_Heap_File, UE_Error, 'Invalid heap file', '' ) ; // Not a file heap file
        Set_Last_Error( Result ) ;
        exit ;
    end ;
    _File_Store._File.Read( 0, 0, sizeof( FH_Header ), FH_Header ) ;
    if(
        ( FH_Header.Prefix <> $FF )
        or
        ( FH_Header.Facility <> Err_UOS_File_Heap_Facility )
      ) then
    begin // UNTESTED
        Result := Create_Simple_UE( Err_UOS_File_Heap_Facility, 10, Err_Invalid_Heap_File, UE_Error, 'Invalid heap file', '' ) ; // Not a file heap file
        Set_Last_Error( Result ) ;
        exit ;
    end ;
    if( FH_Header.Version >= 1 ) then // We only understand file versions < V1.0
    begin // UNTESTED
        Result := Create_Simple_UE( Err_UOS_File_Heap_Facility, 10, Err_Incompatible_Version, UE_Error, 'Invalid heap file', '' ) ;
        Set_Last_Error( Result ) ;
        exit ;
    end ;
    if( ( FH_Header.Resolution < 2 ) or not Power_Of_2( FH_Header.Resolution ) ) then
    begin // UNTESTED
        Result := Create_Simple_UE( Err_UOS_File_Heap_Facility, 10, Err_Invalid_Heap_File, UE_Error, 'Invalid heap file', '' ) ; // Not a file heap file
        Set_Last_Error( Result ) ;
        exit ;
    end ;
    if(
        ( ( FH_Header.AT_Offset and ( FH_Header.Resolution - 1 ) ) <> 0 ) // AT not aligned on resolution boundary
        or
        ( FH_Header.AT_Offset + FH_Header.Resolution > _File_Store._File.File_Size ) // AT beyond EOF
        or
        ( FH_Header.AT_Offset < FH_Header.Resolution ) // AT before the headers
      ) then
    begin
        Result := Create_Simple_UE( Err_UOS_File_Heap_Facility, 10, Err_Invalid_Heap_File, UE_Error, 'Invalid heap file', '' ) ; // Not a file heap file
        Set_Last_Error( Result ) ;
        exit ;
    end ;

    // Load the AT...
    Resolution := FH_Header.Resolution ;
    S := _File_Store._File.File_Size div Resolution ; // Number of clusters in file
    I := ( S + 7 ) div 8 ; // Number of bytes needed to represent the clusters
    Allocation_Table_Offset := FH_Header.AT_Offset ;
    AT.Store := _File_Store ;
    AT.Size := I ;
    AT.Resolution := Resolution ;
    AT.Load_Table ;
    Set_Last_Error( AT.Last_Error ) ;
end ; // TUOS_File_Heap.Init_Heap


// Property handlers...

function TUOS_File_Heap.Get_Resolution : longint ;

begin
    Result := _Resolution ;
end ;


procedure TUOS_File_Heap.Set_Resolution( Value : longint ) ;

var I : longint ;

begin
    I := 4 ; // Minimum resolution
    while( I > 0 ) do
    begin
        if( I = Value ) then
        begin
            _Resolution := I ;
            _File_Store._Resolution := I ;
            exit ;
        end ;
        I := I + I ;
    end ;
end ;


function TUOS_File_Heap.Get_File : TUOS_File ;

begin
    Result := _File_Store._File ;
end ;


function TUOS_File_Heap.Round_Up( S, Resolution : longint ) : longint ;

begin
    Result := ( S + Resolution - 1 ) and not ( Resolution - 1 ) ;
end ;


function TUOS_File_Heap.Get_Origin : int64 ;

var FH_Header : TFH_Header ;

begin
    Set_Last_Error( nil ) ;
    Result := 0 ;
    if( _File_Store._File.File_Size < sizeof( FH_Header ) ) then // File is smaller than the file header
    begin
        Set_Last_Error( Create_Simple_UE( Err_UOS_File_Heap_Facility, 10, Err_Invalid_Heap_File, UE_Error, 'Invalid heap file', '' ) ) ; // Not a file heap file
        exit ;
    end ;
    _File_Store._File.Read( 0, 0, sizeof( FH_Header ), FH_Header ) ;
    Set_Last_Error( _File_Store._File.Last_Error ) ;
    if( Last_Error <> nil ) then
    begin
        exit ;
    end ;
    Result := FH_Header.Origin ;
end ;


procedure TUOS_File_Heap.Set_Origin( Value : int64 ) ;

var FH_Header : TFH_Header ;

begin
    Set_Last_Error( nil ) ;
    if( _File_Store._File.File_Size < sizeof( FH_Header ) ) then // File is smaller than the file header
    begin
        Set_Last_Error( Create_Simple_UE( Err_UOS_File_Heap_Facility, 10, Err_Invalid_Heap_File, UE_Error, 'Invalid heap file', '' ) ) ; // Not a file heap file
        exit ;
    end ;
    _File_Store._File.Read( 0, 0, sizeof( FH_Header ), FH_Header ) ;
    Set_Last_Error( _File_Store._File.Last_Error ) ;
    if( Last_Error <> nil ) then
    begin
        exit ;
    end ;
    FH_Header.Origin := Value ;
    _File_Store._File.Write( 0, 0, sizeof( FH_Header ), FH_Header ) ;
    Set_Last_Error( _File_Store._File.Last_Error ) ;
end ;


function TUOS_File_Heap.Get_Fixed : boolean ;

begin
    Result := _File_Store._Fixed ;
end ;


procedure TUOS_File_Heap.Set_Fixed( Value : boolean ) ;

begin
    _File_Store._Fixed := Value ;
end ;


procedure TUOS_File_Heap.Set_File( Value : TUOS_File ) ;

begin
    Assign_File( Value ) ;
end ; // TUOS_File_Heap.Set_File



// Overrides...

function TUOS_File_Heap.Is_Class( Name : PChar ) : boolean ;

begin
    Result := lowercase( Name ) = 'tuos_file_heap' ;
    if( not Result ) then
    begin
        Result := inherited Is_Class( Name ) ;
    end ;
end ;


function TUOS_File_Heap.Read_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : TUnified_Exception ) : TStore_Address64 ;

begin
    Result := 0 ;
    if( not Valid_Range( Address, _Size ) ) then
    begin
        UEC := Create_Simple_UE( Err_UOS_File_Heap_Facility, 10, Err_Illegal_Call, UE_Error, 'Illegal call', '' ) ;
        exit ;
    end ;
    Result := inherited Read_Data( Data, Address, _Size, UEC ) ;
end ;


function TUOS_File_Heap.Write_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : TUnified_Exception ) : TStore_Address64 ;

begin
    Result := 0 ;
    if( not Valid_Range( Address, _Size ) ) then
    begin
        UEC := Create_Simple_UE( Err_UOS_File_Heap_Facility, 10, Err_Illegal_Call, UE_Error, 'Illegal call', '' ) ;
        exit ;
    end ;
    Result := inherited Write_Data( Data, Address, _Size, UEC ) ;
end ;


function TUOS_File_Heap.Get_Name : PChar ;

begin
    Result := PChar( _Name ) ;
end ;


procedure TUOS_File_Heap.Set_Max_Storage( Value : TStore_Address64 ;
    var Res : TUnified_Exception ) ;

begin
end ;


function TUOS_File_Heap.Allocate( Size : TStore_Address64 ) : TStore_Address64 ;

begin
    Result := Getmem( Size ) ;
end ;


function TUOS_File_Heap.Allocate_At( Offset, Size : TStore_Address64 ) : boolean ;

begin
    Result := False ; // use getmem
end ;


procedure TUOS_File_Heap.Copy( Source, Destination : TStore_Address64 ;
    Count : int64 ) ;

begin
    if( not Valid_Range( Source, Count ) ) then
    begin
        exit ; // Err_Access_Error
    end ;
    if( not Valid_Range( Destination, Count ) ) then
    begin
        exit ; // Err_Access_Error
    end ;
    inherited Copy( Source, Destination, Count ) ;
end ;


procedure TUOS_File_Heap.Deallocate( PTR, Size : TStore_Address64 ) ;

begin
    Freemem( PTR ) ;
end ;


procedure TUOS_File_Heap.Fill( PTR : TStore_Address64 ;
    Count : integer ; Value : byte ) ;

begin
    if( not Valid_Range( PTR, Count ) ) then
    begin
        exit ; // Err_Access_Error
    end ;
    inherited Fill( PTR, Count, Value ) ;
end ;


function TUOS_File_Heap.Min_Storage : TStore_Address64 ;

begin
    Result := Resolution ;
end ;


function TUOS_File_Heap.Reallocate( PTR, Old, New : TStore_Address64 ) : TStore_Address64 ;

begin
    Result := 0 ; // Use reallocmem
end ;


// API...

procedure TUOS_File_Heap.Assign_File( Value : TUOS_File ;
    Resolution : integer = 16 ; Initialize : boolean = False ) ;

begin
    if( _File_Store = nil ) then
    begin
        _File_Store := TUOS_Native_File_Store.Create ;
        _File_Store.Resolution := Resolution ;
        _Resolution := Resolution ;
    end ;
    _File_Store._File := Value ;
    Store := _File_Store ;
    if( Initialize ) then
    begin
        Reset ;
    end else
    if( Value <> nil ) then
    begin
        Init_Heap ;
    end ; // if( Value <> nil )
end ;


function TUOS_File_Heap.Open( FS : TUOS_File_System ;
    Name : string ) : TUnified_Exception ;

var Fil : TUOS_File ;

begin
    Fil := FS.Get_File( pchar( Name ) ) ;
    if( Fil = nil ) then
    begin
        Result := FS.Last_Error ;
        exit ;
    end ;
    _Name := Name ;
    Result := nil ;
    Set_File( Fil ) ;
end ;


function TUOS_File_Heap.Getmem( Size : int64 ) : int64 ;

var ATO : int64 ;

begin
    ATO := Allocation_Table_Offset ;
    Size := Size + sizeof( int64 ) ; // Include space for size prefix
    Size := Round_Up( Size, Resolution ) ; // Round up
    Result := inherited Allocate( Size ) ;
    if( Result <> 0 ) then
    begin
        _File.Write( 0, Result, sizeof( int64 ), Size ) ;
        Result := Result + sizeof( int64 ) ; // Point to the data just past the prefix
    end ;
    if( ATO <> Allocation_Table_Offset ) then // File was resized
    begin
        ATO := Allocation_Table_Offset ;
        if( ATO_File_Position <> 0 ) then
        begin
            _File.Write( 0, ATO_File_Position, sizeof( ATO ), ATO ) ;
        end ;
    end ;
end ;


procedure TUOS_File_Heap.Freemem( PTR : int64 ) ;

var L : longint ;

begin
    L := Get_Length( PTR ) ;
    if( not Valid_Range( PTR, L ) ) then
    begin
        exit ; // Err_Access_Error
    end ;
    inherited Deallocate( PTR - sizeof( int64 ), L ) ;
end ;


function TUOS_File_Heap.Reallocmem( PTR, Size : int64 ) : int64 ;

var ATO : int64 ;
    L : longint ;

begin
    ATO := Allocation_Table_Offset ;
    L := Get_Length( PTR ) ;
    if( not Valid_Range( PTR, L ) ) then
    begin
        Result := 0 ;
        exit ; // Err_Access_Error
    end ;
    Size := Size + sizeof( int64 ) ; // Include space for size prefix
    Size := ( Size + Resolution - 1 ) and not ( Resolution - 1 ) ; // Round up
    Result := inherited Reallocate( PTR - sizeof( int64 ), L, Size ) ;
    if( ATO <> Allocation_Table_Offset ) then // File was resized
    begin
        ATO := Allocation_Table_Offset ;
        if( ATO_File_Position <> 0 ) then
        begin
            _File.Write( 0, ATO_File_Position, sizeof( ATO ), ATO ) ;
        end ;
    end ;
end ;



{$IFDEF CC_Regression_Test}
type TUOS_Test_File = class( TUOS_File )
                          public // Constructors and destructors...
                              destructor Destroy ; override ;

                          private // Instance data...
                              _File : file ; // Data stream buffer
                              _Drive : integer ;

                          public // API (overrides)...
                              function Is_Class( Name : PChar ) : boolean ;
                                  override ;

                              function Create_Stream( Name : int64 ) : longint ;
                                  override ;

                              procedure Delete_Stream( Name : int64 ; Index : longint ) ;
                                  override ;

                              function Max_Stream : longint ; override ;

                              function Stream_Name( Index : longint ) : int64 ;
                                  override ;

                              function Read( Stream : longint ;
                                  Position : TStore_Address64 ;
                                  Length : TStore_Size64 ; var Buff ) : TStore_Size64 ;
                                  override ;

                              function Write( Stream : longint ;
                                  Position : TStore_Address64 ;
                                  Length : TStore_Size64 ; var Buff ) : TStore_Size64 ;
                                  override ;

                			        function Get_Stream_Size( Stream : longint ) : TStore_Size64 ;
                                  override ;

                		          procedure Set_Stream_Size( Stream : longint ;
                                  Value : TStore_Size64 ) ; override ;

                              function Get_Contiguous : boolean ; override ;

                              procedure Set_Contiguous( Value : boolean ) ;
                                  override ;

                              function Get_File_Size : int64 ;
                                  override ;

                              procedure Set_File_Size( Value : int64 ) ;
                                  override ;

                              function Read_Only : boolean ; override ;

                              function Write_Only : boolean ; override ;

                              function XSpaceAvail : int64 ; override ;

                              function Rebuild : boolean ;
                      end ; // TUOS_Test_File

destructor TUOS_Test_File.Destroy ;

begin
    closefile( _File ) ;

    inherited Destroy ;
end ;


function TUOS_Test_File.Is_Class( Name : PChar ) : boolean ;

begin
    Result := lowercase( Name ) = 'tuos_test_file' ;
    if( not Result ) then
    begin
        Result := inherited Is_Class( Name ) ;
    end ;
end ;


function TUOS_Test_File.Create_Stream( Name : int64 ) : longint ;

begin
    Result := -1 ;
end ;


procedure TUOS_Test_File.Delete_Stream( Name : int64 ; Index : longint ) ;

begin
end ;


function TUOS_Test_File.Max_Stream : longint ;

begin
    Result := 0 ;
end ;


function TUOS_Test_File.Stream_Name( Index : longint ) : int64 ;

begin
    Result := 0 ;
end ;


function TUOS_Test_File.Read( Stream : longint ;
    Position : TStore_Address64 ;
    Length : TStore_Size64 ; var Buff ) : TStore_Size64 ;

begin
    if( Stream = 0 ) then
    begin
        seek( _File, Position ) ;
        blockread( _File, Buff, Length ) ;
        Result := Length ;
    end else
    begin
        Result := 0 ;
    end ;
end ;


function TUOS_Test_File.Write( Stream : longint ;
    Position : TStore_Address64 ;
    Length : TStore_Size64 ; var Buff ) : TStore_Size64 ;

begin
    if( Stream = 0 ) then
    begin
        seek( _File, Position ) ;
        blockwrite( _File, Buff, Length ) ;
    end else
    begin
        Result := 0 ;
    end ;
end ;


function TUOS_Test_File.Get_Stream_Size( Stream : longint ) : TStore_Size64 ;

begin
    if( Stream = 0 ) then
    begin
        Result := File_Size ;
    end else
    begin
        Result := 0 ;
    end ;
end ;


procedure TUOS_Test_File.Set_Stream_Size( Stream : longint ;
    Value : TStore_Size64 ) ;

begin
    if( Stream = 0 ) then
    begin
        Set_File_Size( Value ) ;
    end ;
end ;


function TUOS_Test_File.Get_Contiguous : boolean ;

begin
    Result := False ;
end ;


procedure TUOS_Test_File.Set_Contiguous( Value : boolean ) ;
                   
begin
end ;


function TUOS_Test_File.Get_File_Size : int64 ;

begin
    Result := filesize( _File ) ;
end ;


procedure TUOS_Test_File.Set_File_Size( Value : int64 ) ;

var Buffer : array[ 0..255 ] of byte ;
    I : integer ;

begin
    fillchar( Buffer, sizeof( Buffer ), 0 ) ;
    I := Value - filesize( _File ) ;
    seek( _File, filesize( _File ) ) ;
    while( I > 0 ) do
    begin
        if( I < sizeof( Buffer ) ) then
        begin
            blockwrite( _File, Buffer, I ) ;
        end else
        begin
            blockwrite( _File, Buffer, sizeof( Buffer ) ) ;
        end ;
        I := I - sizeof( Buffer ) ;
    end ;
end ;


function TUOS_Test_File.Read_Only : boolean ;

begin
    Result := False ;
end ;


function TUOS_Test_File.Write_Only : boolean ;

begin
    Result := False ;
end ;


function TUOS_Test_File.XSpaceAvail : int64 ;

begin
    Result := DiskFree( _Drive ) ;
end ;


function TUOS_Test_File.Rebuild : boolean ;

begin
    Result := False ;
end ;



procedure Test_Unit( T : TTest_Interface ) ;

    function Log( const S : string ) : boolean ;

    begin
        Last_Op := S ;
        append( F ) ;
        writeln( F, S ) ;
        closefile( F ) ;
        Result := T.Log_Operation( S ) ;
    end ;

procedure Error ;

begin
    halt ;
end ;


var Test_File : TUOS_Test_File ;
    File_Heap : TUOS_File_Heap ;
    I, I1 : int64 ;

begin
    Test_File := TUOS_Test_File.Create ;
    assignfile( Test_File._File, 'c:\work\uos\test.fh' ) ;
    rewrite( Test_File._File, 1 ) ;
    Test_File._Drive := 3 ; // C:

    File_Heap := TUOS_File_Heap.Create ;
    File_Heap.Free ;
    File_Heap := TUOS_File_Heap.Create ;
    File_Heap._File := Test_File ;

    if( File_Heap.Min_Storage <> 16 ) then
    begin
        Error ;
    end ;
    if( File_Heap.Valid_Range( 0, 10 ) ) then
    begin
        Error ;
    end ;
    File_Heap.Set_Resolution( 111 ) ;
    if( File_Heap.Resolution <> 16 ) then
    begin
        Error ;
    end ;
    I := File_Heap.Getmem( 1024 ) ;
    if( I = 0 ) then
    begin
        Error ;
    end ;
    I1 := File_Heap.Getmem( 1024 ) ;
    if( I1 = 0 ) then
    begin
        Error ;
    end ;
    if( I1 < I + 1024 ) then
    begin
        Error ;
    end ;
    File_Heap.Freemem( I1 ) ;
    File_Heap.Freemem( I ) ;
    File_Heap._File := Test_File ;
    File_Heap.Free ;
end ;
{$ENDIF}


begin
//    Test_Unit ;
end.
