/*
        Program Name : _GenVal
        Package Name : Subroutine Library - Layer 0
        Purpose      : Generic Value class definitions.
        Institution  : Conroy & Conroy Co.
        Date Written : 14-February-2005
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2005-2021 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TGeneric_Value is a class that provides a generic value (ie variant).
        There are three intended uses for this (and associated) classes:
          1) RTTI support
          2) Compiler/debugger symbol support
          3) Uber-variant
*/

#ifndef _GENVAL_HPP
	#define _GENVAL_HPP

#include "eDefines.hpp"
#include "_Common.hpp"
#include "COMInter.hpp"
#include "_Types.hpp"
#include <string>

using namespace std;

enum TData_Type_Family { DTF_Unknown, // Variant
                         DTF_Undefined, // Binary
                         DTF_Logical,
                         DTF_Numeric,
                         DTF_Character, // Character/string
                         DTF_Pointer, // pointer/method/procedure/function
                         DTF_Composite,
                         DTF_Collection,
                         DTF_User_Defined,
                         DTF_Meta, // Type
                         DTF_Other
                       } ;

enum TValue_Data_Type { VDT_Empty,
                        VDT_Null,
                        VDT_Boolean,
                        VDT_Integer,
                        VDT_Real,
                        VDT_String,
                        VDT_Char,
                        VDT_Pointer,
                        VDT_Value, // TGeneric_Value defining a type
                        VDT_Code, // Function or procedure
                        VDT_Scalar,
                        VDT_User_Defined,
                        VDT_Unknown
                      } ;

enum TCall_Mechanism { CM_By_Value,
                       CM_By_Reference,
                       CM_By_Name
                     } ;

enum TMethod_Class { MC_Private,
                     MC_Protected,
                     MC_Public,
                     MC_Published
                   } ;

class TGeneric_Value ;

class TGeneric_Parameter: public TDelphi_Object // Parameter information...
{
	
	public:
        virtual TGeneric_Value* Typ() = 0 ; // Type of parameter
        virtual bool Output() = 0 ; // Purpose of parameter is for returning data
        virtual bool Constant() = 0 ; // True if constant
        virtual TCall_Mechanism Mechanism() = 0 ;
        virtual TGeneric_Value* _Default() = 0 ; // Default value of parameter
} ;

class TGeneric_Number: public TDelphi_Object
{
    public:
        virtual bool Fixed() = 0 ; // True if integer or fixed length real
        virtual bool Signed() = 0 ;
        virtual int32 Mantissa_Size() = 0 ; // In bits, for Floating types
        virtual int32 Exponent_Size() = 0 ; // In bits, for Floating types (0 for non-floating)
        virtual int32 Fractional_Size() = 0 ; // In bits, for fixed-point types (0 for non-fixed).
} ;

class TGeneric_Character: public TDelphi_Object
{
    public:
        virtual int32 Max_Length() = 0 ;
        // Maximum allowable length (same as Length except for dynamic-length strings)
        virtual bool Fixed() = 0 ; // True if fixed length
        virtual int32 Length() = 0 ; // Length in bytes
        virtual int32 Logical_Length() = 0 ; // Length in characters
} ;

class TGeneric_Collection: public TDelphi_Object
{
    public:
        virtual bool Fixed() = 0 ; // False if dynamic array 
        virtual bool Sparse() = 0 ; // True if sparse array
        virtual int32 Subscripts() = 0 ; // Number of subscripts
        virtual TGeneric_Value* Low_Bound( int32 Subscript ) = 0 ;
        // Lower bound of subscript for array
        virtual TGeneric_Value* High_Bound( int32 Subscript ) = 0 ;
        // Upper bound of subscript for array
        virtual TGeneric_Value* Get_Subscript( int32 Address ) = 0 ;
        // Get value from array (viewing it as a flat one-dimensional array)
        virtual bool Row_Major() = 0 ; // True if array is left-to-right major
} ;

class TGeneric_Composite: public TDelphi_Object
{
    public:
        virtual TGeneric_Value* Ancestor( int32 index ) = 0 ;
        virtual int32 Ancestor_Count() = 0 ; // Number of ancestors
        virtual TMethod_Class Ancestor_Class( int32 Index ) = 0 ;
        // How Indexth ancestor was inherited

        virtual int32 Count() = 0 ; // Number of values
        virtual TGeneric_Value* Get_Item( int32 Index ) = 0 ; // Return value Index
} ;

class TGeneric_Method: public TDelphi_Object
{
    public:
        virtual bool Is_Abstract() = 0 ; // True if abstract (pure virtual)
        virtual bool Is_Constructor() = 0 ;
        virtual bool Is_Destructor() = 0 ;
        virtual bool Is_Virtual() = 0 ; // True if virtual method
        virtual bool Static() = 0 ; // True if a static class method
        virtual bool C_Order() = 0 ; // True if C calling convention parameter order
        virtual bool Standard() = 0 ; // True if standard (stack) parameter passing
        virtual TGeneric_Value* Execute( TGeneric_Value* Parameters ) = 0 ;
        // Execute the method
        virtual bool Compile_Time() = 0 ;
        // True if the routine can be evaluated at compile-time.
        virtual int32 Count() = 0 ; // Number of parameters (run-time)
        virtual int32 Max_Count() = 0 ; // Maximum potential number of parameters
        virtual TGeneric_Value* Parameter( int Index ) = 0 ;
        // Returns parameter for specified index
        virtual bool Target() = 0 ; // True if a target function
        virtual TMethod_Class Method_Class() = 0 ;
} ;

class TProperty_Information: public TDelphi_Object
{
    public:
        virtual TGeneric_Value* Getter() = 0 ;
        virtual TGeneric_Value* Setter() = 0 ;
        virtual bool _Default() = 0 ;
} ;

class TType_Information: public TDelphi_Object
{
    public:
        virtual bool Is_Composite() = 0 ;
        // True if an array/queue/list/stack of Data_Type
        virtual bool Is_Collection() = 0 ; // True if an object or record
        virtual TMemory_Type Memory_Type() = 0 ;

        virtual TGeneric_Parameter* Parameter() = 0 ;
        // nil if not a parameter
        virtual TGeneric_Number* Numeric_Information() = 0 ;
        // nil if not numeric
        virtual TGeneric_Character* Character_Information() = 0 ;
        // nil if not character/string
        virtual TGeneric_Composite* Composite_Information() = 0 ;
        // nil if not an object/record
        virtual TGeneric_Collection* Collection_Information() = 0 ;
        // nil if not a composite (eg array)
        virtual TGeneric_Method* Method() = 0 ; // nil if not a method/procedure
        virtual TProperty_Information* _Property() = 0  ; // nil if not a property
        virtual int32 Address() = 0 ; // 0 if undefined, type, or parameter

        virtual bool Range_Limited() = 0 ;
        /* True if value is limited to a range of values.
           This is always true for scalars. */

        // (These return nil if not range limited)
        virtual TGeneric_Value* Low_Range() = 0 ;
        virtual TGeneric_Value* High_Range() = 0 ;

            // For scalars...
        virtual string Value_Name( int32 Value ) = 0 ;
        /* Name for given ordinal value.  Returns null
           string if not a scalara or if Value is out of range. */
} ;

class TGeneric_Value: public TDelphi_Object
{
    public: // API...
        // Meta type information...
        virtual TData_Type_Family Family() = 0 ;
        virtual TValue_Data_Type Data_Type() = 0 ;
        virtual string User_Data_Type_Name() = 0 ;
        virtual string Name() = 0 ;
        // Variable, type, method, or parameter name
        virtual TType_Information* Type_Information() = 0 ;

        // Size, address, and format information...
        virtual bool Constant() = 0 ;
        // True if a constant rather than variable
        virtual int32 Size() = 0 ;
        // Size in bytes
        virtual int32 Total_Size() = 0 ; /* Size of all associated data, in bytes.  For
                                instance, Size may return the size of an object
                                but Total_Size also includes the size of any
                                associated data. */

                              // For pointers...
        virtual TGeneric_Value* Dereference() = 0 ;

                              // Conversions/promotions/casts...
        virtual bool Can_Convert( TValue_Data_Type Typ ) = 0 ;
        // True if value can be converted to the specified data type
        virtual TGeneric_Value* Convert( TGeneric_Value* Class_Type, bool &Loss ) = 0 ;

        /* Return a version of the value converted to the
           passed type specification.  Returns nil if the
           conversion cannot be done.  Loss is ignored when
           passed and is True on return if there was a loss of precision. */

        virtual TGeneric_Value* _Clone() = 0 ;

        virtual void Simple_Cast( TValue_Data_Type Typ ) = 0 ;
        virtual void Cast( TGeneric_Value* Typ ) = 0 ;

        // Other...
        virtual void Deserialize( char* Value ) = 0 ;
        virtual char* Serialize() = 0 ;

        // Get/Set value methods...
        virtual void Set_Null() = 0 ;
        virtual void Get_Undefined( void* Buffer, int32 Len ) = 0 ;
        virtual void Set_Undefined( void* Buffer, int32 Len ) = 0 ;

        virtual bool Get_Boolean() = 0 ;
        virtual void Set_Boolean( bool B ) = 0 ;

        virtual ByteBool Get_ByteBool() = 0 ;
        virtual void Set_ByteBool( ByteBool B ) = 0 ;

        virtual WordBool Get_WordBool() = 0 ;
        virtual void Set_WordBool( WordBool B ) = 0 ;

        virtual BOOL Get_LongBool() = 0 ;
        virtual void Set_LongBool( BOOL B ) = 0 ;

        virtual char Get_Char() = 0 ;
        virtual void Set_Char( char C ) = 0 ;

        virtual string Get_String() = 0 ;
        virtual void Set_String( const string S ) = 0 ;

        virtual char* Get_PChar() = 0 ;
        virtual void Set_PChar( char* P ) = 0 ;

        virtual int8 Get_Integer8() = 0 ;
        virtual void Set_Integer8( int8 I ) = 0 ;

        virtual int16 Get_Integer16() = 0 ;
        virtual void Set_Integer16( int16 I ) = 0 ;

        virtual int32 Get_Integer32() = 0 ;
        virtual void Set_Integer32( int32 I ) = 0 ;

        virtual int64 Get_Integer64() = 0 ;
        virtual void Set_Integer64( int64 I ) = 0 ;

        virtual uint8 Get_Unsigned_Integer8() = 0 ;
        virtual void Set_Unsigned_Integer8( uint8 I ) = 0 ;

        virtual uint16 Get_Unsigned_Integer16() = 0 ;
        virtual void Set_Unsigned_Integer16( uint16 I ) = 0 ;

        virtual uint32 Get_Unsigned_Integer32() = 0 ;
        virtual void Set_Unsigned_Integer32( uint32 I ) = 0 ;

        virtual uint64 Get_Unsigned_Integer64() = 0 ;
        virtual void Set_Unsigned_Integer64( uint64 I ) = 0 ;
                                      
        virtual float Get_Single() = 0 ;
        virtual void Set_Single( float S ) = 0 ;

        virtual double Get_Double() = 0 ;
        virtual void Set_Double( double D ) = 0 ;

        virtual long double Get_Extended() = 0 ;
        virtual void Set_Extended( long double E ) = 0 ;

    public: // Properties...
        bool Bool(){return Get_Boolean();};
        ByteBool Byte_Bool(){return Get_ByteBool();};
        WordBool Word_Bool(){return Get_WordBool();};
        BOOL Long_Bool(){return Get_LongBool();};
        char _Char(){return Get_Char();};
        string _String(){return Get_String();};
        char* _PChar(){return Get_PChar();};
        int8 Integer8(){return Get_Integer8();};
        uint8 Unsigned_Integer8(){return Get_Unsigned_Integer8();};
        int16 Integer16(){return Get_Integer16();};
        uint16 Unsigned_Integer16(){return Get_Unsigned_Integer16();};
        int32 Integer32(){return Get_Integer32();};
        uint32 Unsigned_Integer32(){return Get_Unsigned_Integer32();};
        int64 Integer64(){return Get_Integer64();};
        uint64 Integer64(){return Get_Unsigned_Integer();};
        float _Single(){return Get_Single();};
        double _Double(){return Get_Double();};
        long double _Extended(){return Get_Extended();};
} ; // TGeneric_Value

  #ifdef STDCALL_SUPPORTED
class TCOM_Generic_Value;
class TCOM_Generic_Parameter: public TCommon_COM_Interface // Parameter information...
{
    public:
        virtual TCOM_Generic_Value* Typ() = 0 ; // Type of parameter
        virtual bool Output() = 0 ; // Purpose of parameter is for returning data
        virtual bool Constant() = 0 ; // True if constant
        virtual TCall_Mechanism Mechanism() = 0 ;
        virtual TCOM_Generic_Value* _Default() = 0 ;
        // Default value of parameter
} ;

class TCOM_Generic_Number: public TCommon_COM_Interface
{
    public:
        virtual bool Fixed() = 0 ; // True if integer or fixed length real
        virtual bool Signed() = 0 ;
        virtual int32 Mantissa_Size() = 0 ; // In bits, for Floating types
        virtual int32 Exponent_Size() = 0 ;
        // In bits, for Floating types (0 for non-floating)
        virtual int32 Fractional_Size() = 0 ; // In bytes, for fixed-point types
} ;

class TCOM_Generic_Character: public TCommon_COM_Interface
{
    public:
        virtual int32 Max_Length() = 0 ;
        // Maximum allowable length (same as Length except for dynamic-length strings)
        virtual bool Fixed() = 0 ; // True if fixed length
        virtual int32 Length() = 0 ; // Length in bytes
        virtual int32 Logical_Length() = 0 ; // Length in characters
} ;

class TCOM_Generic_Collection: public TCommon_COM_Interface
{
    public:
        virtual bool Fixed() = 0 ; // False if dynamic array
        virtual bool Sparse() = 0 ; // True if sparse array
        virtual int32 Subscripts() = 0 ; // Number of subscripts
        virtual TCOM_Generic_Value* Low_Bound( int32 Subscript ) = 0 ;
        // Lower bound of subscript for array
        virtual TCOM_Generic_Value* High_Bound( int32 Subscript ) = 0 ;
        // Upper bound of subscript for array
        virtual TCOM_Generic_Value* Get_Subscript( int32 Address ) = 0 ;
        // Get value from array (viewing it as a flat one-dimensional array)
        virtual bool Row_Major() = 0 ; // True if array is left-to-right major
} ;

class TCOM_Generic_Composite: public TCommon_COM_Interface
{
    public:
        virtual TCOM_Generic_Value* Ancestor( int32 index ) = 0 ;
        virtual int32 Ancestor_Count() = 0 ; // Number of ancestors
        virtual TMethod_Class Ancestor_Class( int32 Index ) = 0 ;
        // How Indexth ancestor was inherited

        virtual int32 Count() = 0 ; // Number of values
        virtual TCOM_Generic_Value* Get_Item( int32 Index ) = 0 ;
        // Return value Index
} ;

class TCOM_Generic_Method: public TCommon_COM_Interface
{
    public:
        virtual bool Is_Abstract() = 0 ;
        // True if abstract (pure virtual)
        virtual bool Is_Constructor() = 0 ;
        virtual bool Is_Destructor() = 0 ;
        virtual bool Is_Virtual() = 0 ; // True if virtual method
        virtual bool Static() = 0 ; // True if a static class method
        virtual bool C_Order() = 0 ; // True if C calling convention parameter order
        virtual bool Standard() = 0 ; // True if standard (stack) parameter passing
        virtual TCOM_Generic_Value* Execute( TCOM_Generic_Value* Parameters ) = 0 ;
        // Execute the method
        virtual bool Compile_Time() = 0 ;
        // True if the routine can be evaluated at compile-time.
        virtual int32 Count() = 0 ; // Number of parameters (run-time)
        virtual int32 Max_Count() = 0 ; // Maximum potential number of parameters
        virtual TCOM_Generic_Value* Parameter( int Index ) = 0 ;
        // Returns parameter for specified index
        virtual bool Target() = 0 ; // True if a target function
        virtual TMethod_Class Method_Class() = 0 ;
} ;

class TCOM_Property_Information: public TCommon_COM_Interface
{
    public:
	    virtual TCOM_Generic_Value* Getter() = 0 ;
        virtual TCOM_Generic_Value* Setter() = 0 ;
        virtual bool _Default() = 0 ;
} ;

class TCOM_Type_Information: public TCommon_COM_Interface
{
    public:
        virtual bool Is_Composite() = 0 ;
        // True if an array/queue/list/stack of Data_Type
        virtual bool Is_Collection() = 0 ;
        // True if an object or record
        virtual TMemory_Type Memory_Type() = 0 ;

        virtual TCOM_Generic_Parameter* Parameter() = 0 ;
        // nil if not a parameter
        virtual TCOM_Generic_Number* Numeric_Information() = 0 ;
        // nil if not numeric
        virtual TCOM_Generic_Character* Character_Information() = 0 ;
        // nil if not character/string
        virtual TCOM_Generic_Composite* Composite_Information() = 0 ;
        // nil if not an object/record
        virtual TCOM_Generic_Collection* Collection_Information() = 0 ;
        // nil if not a composite (eg array)
        virtual TCOM_Generic_Method* Method() = 0 ;
        // nil if not a method/procedure
        virtual TCOM_Property_Information* _Property() = 0 ;
        // nil if not a property
        virtual int32 Address() = 0 ;
        // 0 if undefined, type, or parameter

        virtual bool Range_Limited() = 0 ;
        /* True if value is limited to a range of values.
           This is always true for scalars. */

        // (These return nil if not range limited)
        virtual TCOM_Generic_Value* Low_Range() = 0 ;
        virtual TCOM_Generic_Value* High_Range() = 0 ;

        // For scalars...
        virtual string Value_Name( int32 Value ) = 0 ;
        /* Name for given ordinal value.  Returns null
           string if not a scalara or if Value is out of range. */
} ;

class TCOM_Generic_Value: public TCommon_COM_Interface
{
    public: // API...
        // Meta type information...
        virtual TData_Type_Family Family() = 0 ;
        virtual TValue_Data_Type Data_Type() = 0 ;
        virtual string User_Data_Type_Name() = 0 ;
        virtual string Name() = 0 ;
        // Variable, type, method, or parameter name
        virtual TCOM_Type_Information* Type_Information() = 0 ;

        // Size, address, and format information...
        virtual bool Constant() = 0 ;
        // True if a constant rather than variable
        virtual int32 Size() = 0 ; // Size in bytes
        virtual int32 Total_Size() = 0 ;
        /* Size of all associated data, in bytes.  For
           instance, Size may return the size of an object
           but Total_Size also includes the size of any
           associated data. */

        // For pointers...
        virtual TCOM_Generic_Value* Dereference() = 0 ;

        // Conversions/promotions/casts...
        virtual bool Can_Convert( TValue_Data_Type Typ ) = 0 ;
        // True if value can be converted to the specified data type
        virtual TCOM_Generic_Value* Convert( TCOM_Generic_Value* Class_Type,
            bool &Loss ) = 0 ;
        /* Return a version of the value converted to the
           passed type specification.  Returns nil if the
           conversion cannot be done.  Loss is ignored
           when passed and is True on return if there
           was a loss of precision. */

        virtual TCOM_Generic_Value* _Clone() = 0 ;

        virtual void Simple_Cast( TValue_Data_Type Typ ) = 0 ;
        virtual void Cast( TCOM_Generic_Value* Typ ) = 0 ;

        // Other...
        virtual void Deserialize( char* Value ) = 0 ;
        virtual char* Serialize() = 0 ;
                                      
        // Get/Set value methods...
        virtual void Set_Null() = 0 ;
        virtual void Get_Undefined( void* Buffer, int32 Len ) = 0 ;
        virtual void Set_Undefined( void* Buffer, int32 Len ) = 0 ;

        virtual bool Get_Boolean() = 0 ;
        virtual void Set_Boolean( bool B ) = 0 ;
                                      
        virtual ByteBool Get_ByteBool() = 0 ;
        virtual void Set_ByteBool( ByteBool B ) = 0 ;

        virtual WordBool Get_WordBool() = 0 ;
        virtual void Set_WordBool( WordBool B ) = 0 ;

        virtual BOOL Get_LongBool() ;
        virtual void Set_LongBool( BOOL B ) = 0 ;


        virtual char Get_Char() = 0 ;
        virtual void Set_Char( char C ) = 0 ;
                                      
        virtual string Get_String() = 0 ;
        virtual void Set_String( const string S ) = 0 ;
                                      
        virtual char* Get_PChar() = 0 ;
        virtual void Set_PChar( char* P ) = 0 ;

        virtual int8 Get_Integer8() = 0 ;
        virtual void Set_Integer8( int8 I ) = 0 ;

        virtual int16 Get_Integer16() = 0 ;
        virtual void Set_Integer16( int16 I ) = 0 ;

        virtual int32 Get_Integer32 () = 0 ;
        virtual void Set_Integer32( int32 I ) = 0 ;

        virtual int64 Get_Integer64() = 0 ;
        virtual void Set_Integer64( int64 I ) = 0 ;

        virtual uint8 Get_Unsigned_Integer8 () = 0 ;
        virtual void Set_Unsigned_Integer8( uint8 I ) = 0 ;

        virtual uint16 Get_Unsigned_Integer16() = 0 ;
        virtual void Set_Unsigned_Integer16( uint16 I ) = 0 ;

        virtual uint32 Get_Unsigned_Integer32() = 0 ;
        virtual void Set_Unsigned_Integer32( uint32 I ) = 0 ;

        virtual uint64 Get_Unsigned_Integer64() = 0 ;
        virtual void Set_Unsigned_Integer64( uint64 I ) = 0 ;

        virtual float Get_Single() = 0 ;
        virtual void Set_Single( float S ) = 0 ;

        virtual double Get_Double() = 0 ;
        virtual void Set_Double( double D ) = 0 ;

        virtual long double Get_Extended() = 0 ;
        virtual void Set_Extended( long double E ) = 0 ;

        bool Bool(){return Get_Boolean();};
        ByteBool Byte_Bool(){return Get_ByteBool();};
        WordBool Word_Bool(){return Get_WordBool();};
        BOOL Long_Bool(){return Get_LongBool();};
        char _Char(){return Get_Char();};
        string _String(){return Get_String();};
        char* _PChar(){return Get_PChar();}; 
        int8 Integer8(){return Get_Integer8();};
        uint8 Unsigned_Integer8(){return Get_Unsigned_Integer8();};
        int16 Integer16(){return Get_Integer16();};
        uint16 Unsigned_Integer16(){return Get_Unsigned_Integer16();};
        int32 Integer32(){return Get_Integer32();};
        uint32 Unsigned_Integer32(){return Get_Unsigned_Integer32();};
        int64 Integer64(){return Get_Integer64();};
        uint64 Unsigned_Integer64(){return Get_Unsigned_Integer64();};
        float _Single(){return Get_Single();};
        double _Double(){return Get_Double();};
        long double _Extended(){return Get_Extended();};
} ;
    #endif

class TSGeneric_Value;
#define PSGeneric_Value TSGeneric_Value*
class TSGeneric_Parameter: public TObject // Parameter information...
{
    public:
        virtual PSGeneric_Value Typ() ; // Type of parameter
        virtual bool Output() ; // Purpose of parameter is for returning data
        virtual bool Constant() ; // True if constant
        virtual TCall_Mechanism Mechanism() ;
        virtual PSGeneric_Value _Default() ; // Default value of parameter
} ;
#define PSGeneric_Parameter TSGeneric_Parameter*

class TSGeneric_Number: public TObject
{
    public:
        virtual bool Fixed() ; // True if integer or fixed length real
        virtual bool Signed() ;
        virtual int32 Mantissa_Size() ; // In bits, for Floating types
        virtual int32 Exponent_Size() ; // In bits, for Floating types (0 for non-floating)
        virtual int32 Fractional_Size() ; // In bytes, for fixed-point types
} ;
#define PSGeneric_Number TSGeneric_Number*

class TSGeneric_Character: public TObject
{
    public:
        virtual int32 Max_Length() ; // Maximum allowable length (same as Length except for dynamic-length strings)
        virtual bool Fixed() ; // True if fixed length
        virtual int32 Length() ; // Length in bytes
        virtual int32 Logical_Length() ; // Length in characters
} ;
#define PSGeneric_Character TSGeneric_Character*

class TSGeneric_Collection: public TObject
{
    public:
        virtual bool Fixed() ; // False if dynamic array
        virtual bool Sparse() ; // True if sparse array
        virtual int32 Subscripts() ; // Number of subscripts
        virtual PSGeneric_Value Low_Bound( int32 Subscript ) ;
        // Lower bound of subscript for array
        virtual PSGeneric_Value High_Bound( int32 Subscript ) ;
        // Upper bound of subscript for array
        virtual PSGeneric_Value Get_Subscript( int32 Address ) ;
        // Get value from array (viewing it as a flat one-dimensional array)
        virtual bool Row_Major() ; // True if array is left-to-right major
} ;
#define PSGeneric_Collection TSGeneric_Composite*

class TSGeneric_Composite: public TObject
{
    public:
        virtual PSGeneric_Value Ancestor( int32 index )  ;
        virtual int32 Ancestor_Count() ; // Number of ancestors
        virtual TMethod_Class Ancestor_Class( int32 Index ) ;
        // How ancestor was inherited

        virtual int32 Count() ; // Number of values
        virtual PSGeneric_Value Get_Item( int32 Index ) ;
        // Return value Index
} ;
#define PSGeneric_Composite TSGeneric_Composite*

class TSGeneric_Method: public TObject
{
    public:
        virtual bool Is_Abstract() ; // True if abstract (pure virtual)
        virtual bool Is_Constructor() ;
        virtual bool Is_Destructor() ;
        virtual bool Is_Virtual() ; // True if virtual method
        virtual bool Static() ; // True if a static class method
        virtual bool C_Order() ; // True if C calling convention parameter order
        virtual bool Standard() ; // True if standard (stack) parameter passing
        virtual TSGeneric_Method* Execute( TSGeneric_Method* Parameters ) ;
        // Execute the method
        virtual bool Compile_Time() ;
        // True if the routine can be evaluated at compile-time.
        virtual int32 Count() ; // Number of parameters (run-time)
        virtual int32 Max_Count() ; // Maximum potential number of parameters
        virtual PSGeneric_Value Parameter( int Index ) ;
        // Returns parameter for specified index
        virtual bool Target() ; // True if a target function
        virtual TMethod_Class Method_Class() ;
} ;
#define PSGeneric_Method TSGeneric_Method*

class TSProperty_Information: public TObject
{
    public:
        virtual PSGeneric_Value Getter() ; 
        virtual PSGeneric_Value Setter() ; 
        virtual bool _Default() ; 
} ;
#define PSProperty_Information TSProperty_Information*

class TSType_Information: public TObject
{
    public:
	    virtual bool Is_Composite() ; // True if an array/queue/list/stack of Data_Type
        virtual bool Is_Collection() ; // True if an object or record
        virtual TMemory_Type Memory_Type() ; 

        virtual PSGeneric_Parameter Parameter() ;
        // nil if not a parameter
        virtual PSGeneric_Number Numeric_Information() ;
        // nil if not numeric
        virtual PSGeneric_Character Character_Information() ;
        // nil if not character/string
        virtual PSGeneric_Composite Composite_Information() ;
        // nil if not an object/record
        virtual PSGeneric_Collection Collection_Information() ;
        // nil if not a composite (eg array)
        virtual PSGeneric_Method Method() ; 
        // nil if not a method/procedure
        virtual PSProperty_Information _Property() ;
        // nil if not a property
        virtual int32 Address() ; // 0 if undefined, type, or parameter

        virtual bool Range_Limited() ;
        /* True if value is limited to a range of values.
           This is always true for scalars. */

        // (These return nil if not range limited)
        virtual PSGeneric_Value Low_Range() ; 
        virtual PSGeneric_Value High_Range() ; 

        // For scalars...
        virtual string Value_Name( int32 Value ) ;
        /* Name for given ordinal value.  Returns null
           string if not a scalara or if Value is out of range. */
} ;
#define PSType_Information TSType_Information*

class TSGeneric_Value: public TObject
{
    public: // API...
        // Meta type information...
        virtual TData_Type_Family Family() ;
        virtual TValue_Data_Type Data_Type() ;
        virtual string User_Data_Type_Name() ;
        virtual string Name() ; // Variable, type, method, or parameter name
        virtual PSType_Information Type_Information() ;

        // Size, address, and format information...
        virtual bool Constant() ; // True if a constant rather than variable
        virtual int32 Size() ; // Size in bytes
        virtual int32 Total_Size() ;
        /* Size of all associated data, in bytes.  For
           instance, Size may return the size of an object
           but Total_Size also includes the size of any associated data. */

        // For pointers...
        virtual PSGeneric_Value Dereference() ;

        // Conversions/promotions/casts...
        virtual bool Can_Convert( TValue_Data_Type Typ ) ;
        // True if value can be converted to the specified data type

        virtual PSGeneric_Value _Clone() ;

        virtual PSGeneric_Value Convert( PSGeneric_Value Class_Type, bool &Loss ) ;
        /* Return a version of the value converted to the
           passed type specification.  Returns nil if the
           conversion cannot be done.  Loss is ignored
           when passed.  On return it is True if there was
           a loss of precision. */

        virtual void Simple_Cast( TValue_Data_Type Typ ) ;
        virtual void Cast( PSGeneric_Value Typ ) ;

        // Other...
        virtual void Deserialize( char* Value ) ;
        virtual char* Serialize() ;

        // Get/Set value methods...
        virtual void Set_Null() ; 
        virtual void Get_Undefined( void* Buffer, int32 Len ) ;
        virtual void Set_Undefined( void* Buffer, int32 Len ) ; 

        virtual bool Get_Boolean() ; 
        virtual void Set_Boolean( bool B ) ;

        virtual ByteBool Get_ByteBool() ;
        virtual void Set_ByteBool( ByteBool B ) ;

        virtual WordBool Get_WordBool() ;
        virtual void Set_WordBool( WordBool B ) ;

        virtual BOOL Get_LongBool() ;
        virtual void Set_LongBool( BOOL B ) ;

        virtual char Get_Char() ; 
        virtual void Set_Char( char C ) ;

        virtual string Get_String() ; 
        virtual void Set_String( const string S ) ; 

        virtual char* Get_PChar() ;
        virtual void Set_PChar( char* P ) ; 

        virtual int8 Get_Integer8() ; 
        virtual void Set_Integer8( int8 I ) ; 

        virtual int16 Get_Integer16() ; 
        virtual void Set_Integer16( int16 I ) ; 

        virtual int32 Get_Integer32() ; 
        virtual void Set_Integer32( int32 I ) ; 

        virtual int64 Get_Integer64() ;
        virtual void Set_Integer64( int64 I ) ;

        virtual uint8 Get_Unsigned_Integer8() ;
        virtual void Set_Unsigned_Integer8( uint8 I ) ;

        virtual uint16 Get_Unsigned_Integer16() ;
        virtual void Set_Unsigned_Integer16( uint16 I ) ;

        virtual uint32 Get_Unsigned_Integer32() ;
        virtual void Set_Unsigned_Integer32( uint32 I ) ;

        virtual uint64 Get_Unsigned_Integer64() ;
        virtual void Set_Unsigned_Integer64( uint64 I ) ;

        virtual float Get_Single() ;
        virtual void Set_Single( float S ) ; 

        virtual double Get_Double() ; 
        virtual void Set_Double( double D ) ; 

        virtual long double Get_Extended() ; 
        virtual void Set_Extended( long double E ) ; 

        bool Bool(){return Get_Boolean();};
        ByteBool Byte_Bool(){return Get_ByteBool();};
        WordBool Word_Bool(){return Get_WordBool();};
        BOOL Long_Bool(){return Get_LongBool();};
        char _Char(){return Get_Char();};
        string _String(){return Get_String();};
        char* _PChar(){return Get_PChar();};
        int8 Integer8(){return Get_Integer8();};
        uint8 Unsigned_Integer8(){return Get_Unsigned_Integer8();};
        int16 Integer16(){return Get_Integer16();};
        uint16 Unsigned_Integer16(){return Get_Unsigned_Integer16();};
        int32 Integer32(){return Get_Integer32();};
        uint32 Unsigned_Integer32(){return Get_Unsigned_Integer32();};
        int64 Integer64(){return Get_Integer64();};
        uint64 Unsigned_Integer64(){return Get_Unsigned_Integer64();};
        float _Single(){return Get_Single();};
        double _Double(){return Get_Double();};
        long double _Extended(){return Get_Extended();};
} ;

#endif
