unit GUI ;

interface

type TRectangle = packed record
                      L, T, R, B : integer ;
                  end ;

     TPixel_Rec = packed record
                      case boolean of
                          false: ( Red, Green, Blue, Alpha : word ) ;
                          true: ( Color : int64 ) ;
                  end ;

     Tcolor_Rec = packed record
                      case boolean of
                          false: ( Red, Green, Blue, Alpha : byte ) ;
                          true: ( Color : integer )
                  end ;

     TPixel = int64 ; // RGBA (word values)

     TImage = class // PNG image
                  public // API
              end ;

     TCanvas = class // Generic drawing surface
                   public // API...
                       procedure CopyFrom( Source : TCanvas ;
                           Src, Dst : TRectangle ) ;
                           virtual ; abstract ;
               end ;

     TBackground_Type = ( BT_Color, BT_Image, BT_Brush, BT_Gradient, BT_Bigradient ) ;

     TBackground = class
                       public
                           BackgroundType : TBackground_Type ;
                           Color : TPixel ;
                   end ;

     TSurface = class ;

     TPointer = class ;

     TGUI_Object = class
                       protected // Property handlers...
                           function Get_Parent : TGUI_Object ;
                               virtual ; abstract ;
                           procedure Set_Parent( Value : TGUI_Object ) ;
                               virtual ; abstract ;
                           procedure Set_Width( Value : integer ) ;
                               virtual ; abstract ;
                           procedure Set_Height( Value : integer ) ;
                               virtual ; abstract ;
                           function Get_Width : integer ; virtual ; abstract ;
                           function Get_Height : integer ; virtual ; abstract ;

                       public // API...
                           Background : TBackground ;

                           procedure AspectChange( Source : TGUI_Object ) ;
                               virtual ; abstract ;
                           function Canvas : TCanvas ; virtual ; abstract ;
                           procedure Update ;
                               virtual ; abstract ;

                       public // Properties
                           property Parent : TGUI_Object
                               read Get_Parent
                               write Set_Parent ;
                           property Width : integer
                               read Get_Width
                               write Set_Width ;
                           property Height : integer
                               read Get_Height
                               write Set_Height ;
                   end ; // TGUI_Object

     TWorld = class( TGUI_Object ) // 3D world or viewports
                  protected // Property handlers...
                      function Get_Viewport : TSurface ;
                          virtual ; abstract ;
                      procedure Set_Viewport( Value : TSurface ) ;
                          virtual ; abstract ;
                      function Get_Object( Index : integer ) : TSurface ;
                          virtual ; abstract ;
                      procedure Set_Object( Index : integer ; Value : TSurface ) ;
                          virtual ; abstract ;
                      function Get_Mouse : TPointer ; virtual ; abstract ;

                  public // API...
                      procedure Add_Object( Value : TSurface ) ;
                          virtual ; abstract ;
                      procedure BringToFront( Value : TSurface ) ;
                          virtual ; abstract ;
                      procedure Delete_Object( Value : TSurface ) ;
                          virtual ; abstract ;
                      procedure Delete_Object_Index( Index : integer ) ;
                          virtual ; abstract ;
                      procedure Clear ; virtual ; abstract ;
                      function Object_Count : integer ; virtual ; abstract ;
                      function Object_At( X, Y, Z : integer ) : TSurface ;
                          virtual ; abstract ;

                  public // Properties
                      property Viewport : TSurface
                          read Get_Viewport
                          write Set_Viewport ; // Where world is drawn
                      property Objects[ Index : integer ] : TSurface
                          read Get_Object
                          write Set_Object ;
                      property Mouse : TPointer
                          read Get_Mouse ;
              end ;

     TSurface = class( TGUI_Object ) // 2D surface - something to draw upon
                    protected // Property handlers...
                        function Get_World : TWorld ; virtual ; abstract ;
                        procedure Set_World( Value : TWorld ) ; virtual ; abstract ;
                        function Get_Top : integer ; virtual ; abstract ;
                        procedure Set_Top( Value : integer ) ; virtual ; abstract ;
                        function Get_Left : integer ; virtual ; abstract ;
                        procedure Set_Left( Value : integer ) ; virtual ; abstract ;
                        function Get_Pixel( X, Y : integer ) : TPixel ;
                            virtual ; abstract ;
                        procedure Set_Pixel( X, Y : integer ; Value : TPixel ) ;
                            virtual ; abstract ;

                    public // API...
                        procedure BringToFront ; virtual ; abstract ;
                        procedure Draw( const DestR, SrcR : TRectangle ;
                            Source : Timage ) ; virtual ; abstract ;
                        procedure Set_Dimensions( H, W : integer ) ;
                            virtual ; abstract ;
                        procedure FillRect( Top, Left, Right, Bottom : integer ;
                            Background_Color : TPixel ) ;
                            virtual ; abstract ;

                    public // Properties...
                        property Pixels[ X, Y : integer ] : TPixel
                            read Get_Pixel
                            write Set_Pixel ;
                        property World : TWorld
                            read Get_World
                            write Set_World ;
                        property Top : integer
                            read Get_Top
                            write Set_Top ;
                        property Left : integer
                            read Get_Left
                            write Set_Left ;
                end ; // TSurface

     TAnimation = class // Animated image
                      public // API
                          procedure Show( Surface : TSurface ) ; virtual ; abstract ;
                          procedure Stop ; virtual ; abstract ;
                          procedure Hide ; virtual ; abstract ;
                  end ;


     TPointer = class // Mouse (or other) pointer
                    public // API
                         procedure Show ; virtual ; abstract ;
                         procedure Hide ; virtual ; abstract ;
                         procedure Set_Position( X, Y, Z : integer ) ;
                              virtual ; abstract ;
                end ;

type TImage_Pointer = class( TPointer ) // pointer with an image
                          public // API...
                              procedure Set_Blink_Rate ; virtual ; abstract ;
                              function Get_Image : TImage ; virtual ; abstract ;
                              procedure Set_Image( Value : TImage ) ;
                                  virtual ; abstract ;
                      end ;

type TAnimated_Pointer = class( TPointer ) // pointer with an animation
                             public // API...
                                 function Get_Animation : TAnimation ;
                                     virtual ; abstract ;
                                 procedure Set_Animation( Value : TAnimation ) ;
                                     virtual ; abstract ;
                         end ;

function Color32_To_Color64( Clr : integer ) : TPixel ;
function Color64_To_Color32( Clr : int64 ) : integer ;
function Make_Rect( T, L, R, B : integer ) : TRectangle ;


implementation

uses Graphics ; // TColor

function Color32_To_Color64( Clr : integer ) : TPixel ;

var C : TPixel_Rec ;
    S : TColor_Rec ;

begin
    S.Color := Clr ;
    C.Alpha := $FFFF ;
    C.Red := S.Red shl 8 ;
    C.Blue := S.Blue shl 8 ;
    C.Green := S.Green shl 8 ;
    Result := C.Color ;
end ;


function Color64_To_Color32( Clr : int64 ) : integer ;

var C : TPixel_Rec ;
    S : TColor_Rec ;

begin
    C.Color := Clr ;
    S.Alpha := 0 ;
    S.Red := C.Red shr 8 ;
    S.Blue := C.Blue shr 8 ;
    S.Green := C.Green shr 8 ;
    Result := S.Color ;
end ;


function Make_Rect( T, L, R, B : integer  ) : TRectangle ;

begin
    Result.T := T ;
    Result.L := L ;
    Result.R := R ;
    Result.B := B ;
end ;



end.

